#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################
# Interfacing with scikit-learn

from .features import featureScaler
from .sequences import sequences, positive, negative
from .models import sequenceModel
import cupy as cp
import numpy as np
from sklearn import svm
from .featurenetwork import baseModel

# GPU implementation
class CUDASVM(baseModel):
	
	def __init__(self, model = None, labelPositive = positive, labelNegative = negative, kDegree = 1, C = 4, batchsize = 1000):
		super().__init__(enableMultiprocessing = False, batchsize = batchsize)
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.labelNegative = labelNegative
		self.C, self.kDegree = C, kDegree
		self.mdl = model
		if model is not None:
			self.SV = cp.array(model.support_vectors_)
			self.cN = cp.array(model.dual_coef_[0])
	
	def __str__(self):
		return 'Support Vector Machine (CUDA)<C: %s; Kernel degree: %s; Positive label: %s; Negative label: %s>'%(str(self.C), str(self.kDegree), str(self.labelPositive), str(self.labelNegative))
	
	def score(self, featureVectors):
		c0 = self.mdl.coef0
		bias = self.mdl.intercept_[0]
		nfeatures = len(self.mdl.support_vectors_[0])
		gamma = 1.0 / (nfeatures)
		ret = []
		fvs = cp.array(featureVectors)
		rmat = ((gamma * (fvs @ self.SV.T) + c0) ** self.kDegree) @ self.cN + bias
		return [ [ v ] for v in rmat ]
		#return [ [ float(v) ] for v in rmat ]
	
	def train(self, trainingSet):
		fvPos = trainingSet[self.labelPositive]
		fvNeg = trainingSet[self.labelNegative]
		cP = [ 1.0 for _ in range(len(fvPos)) ]
		cN = [ -1.0 for _ in range(len(fvNeg)) ]
		mdl = svm.SVC(C = self.C, kernel = 'poly', degree = self.kDegree, gamma = 'auto')
		mdl.fit( np.array(fvPos+fvNeg), np.array(cP+cN) )
		return CUDASVM(
			model = mdl,
			labelPositive = self.labelPositive,
			labelNegative = self.labelNegative,
			C = self.C,
			kDegree = self.kDegree)
	
	def weights(self, featureNames):
		return nctable(
			'Weights: ' + str(self),
			{
				**{
					'Feature': self.featureNames
				},
				**{
					'Weight': self._weights
				}
			},
			align = { 'Feature': 'l' }
		)
	
	def __repr__(self): return self.__str__()




# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################

from .about import *
from .ioputil import nctable
from .common import setSeed, nucleotides, complementaryNucleotides, getReverseComplementaryDNASequence, IUPACNucleotideCodes, IUPACNucleotideCodeSemantics, mean, std, CI, setConfidenceIntervalFunction, useSciPyConfidenceIntervals, KLdiv
from .regions import region, regions, loadBED, loadBEDGZ, loadGFF, loadGFFGZ, loadCoordinateList, predictionBarplot, overlapSensitivityBarplot, overlapPrecisionBarplot, nucleotidePrecisionBarplot, nucleotideRegionF1Barplot
from .sequences import sequence, sequences, sequenceStream, loadFASTA, loadFASTAGZ, streamFASTA, streamFASTAGZ, stream2bit, streamSequenceWindows, getSequenceWindowRegions, MarkovChain, IID, sequenceLabel, positive, negative
from .curves import fixedStepCurve, variableStepCurveValue, variableStepCurve, curves, loadWIG, loadWIGGZ
from .motifs import motifOccurrence, motifs, IUPACMotif, PWMMotif, loadMEMEPWMDatabase
from .features import feature, features, scaledFeature, featureScaler, featureMotifOccurrenceFrequency, featurePREdictorMotifPairOccurrenceFrequency, kSpectrum, kSpectrumMM, kSpectrumGPS
from .featurenetwork import featureNetworkNode, FNNMotifOccurrenceFrequencies, FNNMotifPairOccurrenceFrequencies, FNNLogOdds, kSpectrum, kSpectrumMM, FNNScaler, FNNWindow, FNNModel, baseModel, logOdds
from .models import setNCores, setNThreadFetch, sequenceModel, sequenceModelDummy, sequenceModelLogOdds, trainSinglePREdictorModel, createDummyPREdictorModel, trainPREdictorModel, crossvalidate
from .biomarkers import biomarkers
from .validation import point2D, validationPair, getROC, getPRC, getAUC, getConfusionMatrix, getConfusionMatrixStatistics, printValidationStatistics
from .genome import gene, genome, plotGenomeTracks




# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################

__name__ = "gnocis"
__version__ = "0.1.0"
__summary__ = (
    "Gnocis is a system for the analysis and the modelling of cis-regulatory DNA sequences."
)
__uri__ = "https://github.com/bjornbredesen/gnocis"
__author__ = "Bjørn Bredesen"
__email__ = "bjorn@bjornbredesen.no"
__license__ = "MIT"
__title__ = "gnocis"
__release__ = True




#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################
# Interfacing with scikit-learn

from .features import featureScaler
from .models import sequenceModel
from .sequences import sequences, positive, negative
import numpy as np
import cupy as cp
from sklearn import svm
from sklearn.ensemble import RandomForestClassifier
import random

# Support Vector Machines
class sequenceModelSVMOptimizedQuadratic(sequenceModel):
	"""
	The `sequenceModelSVMOptimizedQuadratic` class trains a quadratic kernel Support Vector Machine (SVM) using scikit-learn. The kernel is applied using matrix multiplication.
	
	:param name: Model name.
	:param features: Feature set.
	:param positives: Positive training sequences.
	:param negatives: Negative training sequences.
	:param windowSize: Window size.
	:param windowStep: Window step size.
	:param kDegree: Kernel degree.
	
	:type name: str
	:type features: features
	:type positives: sequences
	:type negatives: sequences
	:type windowSize: int
	:type windowStep: int
	:type kDegree: float
	"""
	
	def __init__(self, name, features, trainingSet, windowSize, windowStep, kDegree, scale = True, labelPositive = positive, labelNegative = negative):
		super().__init__(name)
		self.windowSize, self.windowStep = windowSize, windowStep
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.trainingSet = trainingSet
		positives, negatives = trainingSet.withLabel([ labelPositive, labelNegative ])
		wPos = sequences(positives.name, [ w for s in positives for w in s.windows(self.windowSize, self.windowStep) ])
		wNeg = sequences(negatives.name, [ w for s in negatives for w in s.windows(self.windowSize, self.windowStep) ])
		if scale:
			features = featureScaler( features, trainingSet = wPos + wNeg )
		self.scale = scale
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.kernel = kDegree
		self.threshold = 0.0
		assert(kDegree == 2)
		vP = [ self.getSequenceFeatureVector(w) for w in wPos ]
		vN = [ self.getSequenceFeatureVector(w) for w in wNeg ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		cls.fit( np.array(vP+vN), np.array(cP+cN) )
		assert(cls.coef0 == 0.0)
		# Extract basic kernel values
		SV = cls.support_vectors_
		nfeatures = len(cls.support_vectors_[0])
		bias = cls.intercept_[0]
		c0 = cls.coef0
		c = cls.dual_coef_[0]
		gamma = 1.0 / (nfeatures)
		self.nSV = len(cls.support_vectors_)
		# Extract weight matrix
		self.pairWeights = ((SV.T * c) @ SV) * gamma * gamma
		# Extract Bias
		self.bias = -bias
	
	def getTrainer(self):
		return lambda ts: sequenceModelSVMOptimizedQuadratic(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, alpha = self.alpha, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
	
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	
	def scoreWindow(self, seq):
		fv = np.array(self.getSequenceFeatureVector(seq))
		return self.pairWeights @ fv @ fv - self.bias
	
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Training set: %s; Positive label: %s; Negative label; Kernel: %s; Support vectors: %d; Quadratic optimized>'%(str(self.features), len(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], self.nSV)
	
	def __repr__(self): return self.__str__()


# Support Vector Machines
class sequenceModelSVMOptimizedQuadraticAutoScale(sequenceModel):
	"""
	The `sequenceModelSVMOptimizedQuadratic` class trains a quadratic kernel Support Vector Machine (SVM) using scikit-learn. The kernel is applied using matrix multiplication.
	
	:param name: Model name.
	:param features: Feature set.
	:param positives: Positive training sequences.
	:param negatives: Negative training sequences.
	:param windowSize: Window size.
	:param windowStep: Window step size.
	:param kDegree: Kernel degree.
	
	:type name: str
	:type features: features
	:type positives: sequences
	:type negatives: sequences
	:type windowSize: int
	:type windowStep: int
	:type kDegree: float
	"""
	
	def __init__(self, name, features, trainingSet, windowSize, windowStep, kDegree, scale = True, labelPositive = positive, labelNegative = negative):
		super().__init__(name)
		self.windowSize, self.windowStep = windowSize, windowStep
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.trainingSet = trainingSet
		positives, negatives = trainingSet.withLabel([ labelPositive, labelNegative ])
		wPos = sequences(positives.name, [ w for s in positives for w in s.windows(self.windowSize, self.windowStep) ])
		wNeg = sequences(negatives.name, [ w for s in negatives for w in s.windows(self.windowSize, self.windowStep) ])
		if scale:
			features = featureScaler( features, trainingSet = wPos + wNeg )
		self.scale = scale
		self.features = features
		self.kernel = kDegree
		self.threshold = 0.0
		assert(kDegree == 2)
		vP = [ self.getSequenceFeatureVector(w) for w in wPos ]
		vN = [ self.getSequenceFeatureVector(w) for w in wNeg ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		cls.fit( np.array(vP+vN), np.array(cP+cN) )
		assert(cls.coef0 == 0.0)
		# Extract basic kernel values
		SV = cls.support_vectors_
		nfeatures = len(cls.support_vectors_[0])
		bias = cls.intercept_[0]
		c0 = cls.coef0
		c = cls.dual_coef_[0]
		gamma = 1.0 / (nfeatures)
		self.nSV = len(cls.support_vectors_)
		# Extract weight matrix
		alpha = [ ( f.vSub - 1.0 ) / f.vScale for f in features ]
		beta = [ ( f.vSub + 1.0 ) / f.vScale for f in features ]
		#
		self.smat = np.array([
			[
				(1./(beta[i]-alpha[i])) * (1./(beta[j]-alpha[j]))
				for j in range(len(features))
			]
			for i in range(len(features))
		])
		SVProd = ( (SV.T * c) @ SV ) * gamma * gamma
		self.pairWeights = SVProd * 4. * self.smat
		#
		self.linWeights = np.array([
			sum(
				-4. * SVProd[i, j] * self.smat[i, j] * (alpha[j]+beta[j])
				for j in range(len(features))
			)
			for i in range(len(features))
		])
		#
		self.bias = sum(
						gamma * gamma * _c * (sum(
							_v[i] * (alpha[i]+beta[i]) / (beta[i] - alpha[i])
							for i in range(len(features))
						)**2.)
					for _v, _c in zip(SV, c)
				) + cls.intercept_[0]
	
	def getTrainer(self):
		return lambda ts: sequenceModelSVMOptimizedQuadraticAutoScale(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, alpha = self.alpha, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
	
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	
	def scoreWindow(self, seq):
		fraw = nc.features('', [ f.feature for f in self.features ])
		fv = np.array(fraw.getAll(seq))
		return (self.pairWeights @ fv @ fv) + (self.linWeights @ fv) + self.bias
	
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Training set: %s; Positive label: %s; Negative label; Kernel: %s; Support vectors: %d; Quadratic optimized>'%(str(self.features), len(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], self.nSV)
	
	def __repr__(self): return self.__str__()


# Support Vector Machines
class sequenceModelSVMOptimizedQuadraticCUDA(sequenceModel):
	"""
	The `sequenceModelSVMOptimizedQuadraticCUDA` class trains a quadratic kernel Support Vector Machine (SVM) using scikit-learn. The kernel is applied using matrix multiplication with CUDA.
	
	:param name: Model name.
	:param features: Feature set.
	:param positives: Positive training sequences.
	:param negatives: Negative training sequences.
	:param windowSize: Window size.
	:param windowStep: Window step size.
	:param kDegree: Kernel degree.
	
	:type name: str
	:type features: features
	:type positives: sequences
	:type negatives: sequences
	:type windowSize: int
	:type windowStep: int
	:type kDegree: float
	"""
	
	def __init__(self, name, features, trainingSet, windowSize, windowStep, kDegree, scale = True, labelPositive = positive, labelNegative = negative):
		super().__init__(name)
		self.windowSize, self.windowStep = windowSize, windowStep
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.trainingSet = trainingSet
		positives, negatives = trainingSet.withLabel([ labelPositive, labelNegative ])
		wPos = sequences(positives.name, [ w for s in positives for w in s.windows(self.windowSize, self.windowStep) ])
		wNeg = sequences(negatives.name, [ w for s in negatives for w in s.windows(self.windowSize, self.windowStep) ])
		if scale:
			features = featureScaler( features, trainingSet = wPos + wNeg )
		self.scale = scale
		self.features = features
		self.kernel = kDegree
		self.threshold = 0.0
		assert(kDegree == 2)
		vP = [ self.getSequenceFeatureVector(w) for w in wPos ]
		vN = [ self.getSequenceFeatureVector(w) for w in wNeg ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		cls.fit( np.array(vP+vN), np.array(cP+cN) )
		assert(cls.coef0 == 0.0)
		# Extract basic kernel values
		SV = cls.support_vectors_
		nfeatures = len(cls.support_vectors_[0])
		bias = cls.intercept_[0]
		c0 = cls.coef0
		c = cls.dual_coef_[0]
		gamma = 1.0 / (nfeatures)
		self.nSV = len(cls.support_vectors_)
		# Extract weight matrix
		self.pairWeights = cp.array( ((SV.T * c) @ SV) * gamma * gamma )
		# Extract Bias
		self.bias = -bias
	
	def getTrainer(self):
		return lambda ts: sequenceModelSVMOptimizedQuadraticCUDA(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, alpha = self.alpha, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
	
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	
	def scoreWindow(self, seq):
		fv = cp.array(self.getSequenceFeatureVector(seq))
		return self.pairWeights @ fv @ fv - self.bias
	
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Training set: %s; Positive label: %s; Negative label; Kernel: %s; Support vectors: %d; Quadratic optimized, CUDA>'%(str(self.features), len(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], self.nSV)
	
	def __repr__(self): return self.__str__()




#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################

from .regions import *
from .curves import curves, variableStepCurve, variableStepCurveValue, fixedStepCurve
from .sequences import *
from .ioputil import nctable
from .common import mean, CI

# Represents annotated genes
class gene:
	
	def __init__(self, name, region, synonyms = None, exons = None, CDS = None):
		self.name = name
		self.synonyms = synonyms
		self.region = region
		self.exons = [] if exons is None else exons
		self.CDS = [] if CDS is None else CDS
	
	def __len__(self):
		return self.end+1-self.start
	
	def __str__(self):
		rname = '%s (%s:%d..%d (%s))'%(self.name, self.region.seq, self.region.start, self.region.end, '+' if self.region.strand else '-')
		return 'Gene<%s>'%(rname)
	
	def __repr__(self):
		return self.__str__()


# Represents a genome
class genome:
	
	def __init__(self, name, genes = None, sequences = None, chromosomes = None, annotation = None, annotationPath = None, seqPath = None, seqLens = None):
		self.name = name
		self.genes = genes
		self.sequences = sequences
		self.annotation = annotation
		self.chromosomes = set() if chromosomes is None else chromosomes
		self.annotationPath = annotationPath
		self.seqPath = seqPath
		self.seqLens = seqLens
		self.genesByName = {  }
		self.genesBySyn = {  }
		if genes is not None:
			for g in genes:
				self.genesByName[g.name] = g
				for s in g.synonyms:
					self.genesBySyn[s] = g
	
	def loadEnsemblAnnotationGTFGZ(self, path):
		dat = loadGFFGZ(path)
		annotation = dat
		annotationPath = path
		chromosomes = set(r.seq for r in dat)
		genes = dat.filter('Genes', lambda r: r.feature == 'gene')
		exons = dat.filter('Exons', lambda r: r.feature == 'exon')
		CDS = dat.filter('CDS', lambda r: r.feature == 'CDS')
		getAnnoDict = lambda r: {
			div[0].strip(): div[1].strip()[1:-1]
			for div in (
				q for q in (
					part.strip().split(' ')
					for part in r.group.split(';')
				)
				if len(q) >= 2
			)
		}
		genesByName = {  }
		for r in genes:
			_dict = getAnnoDict(r)
			name = _dict['gene_name']
			g = gene(name = name, region = r, synonyms = [ name, _dict['gene_id'] ])
			genesByName[name] = g
		for r in exons:
			_dict = getAnnoDict(r)
			name = _dict['gene_name']
			genesByName[name].exons.append(r)
		for r in CDS:
			_dict = getAnnoDict(r)
			name = _dict['gene_name']
			genesByName[name].CDS.append(r)
		genes = [ genesByName[k] for k in genesByName ]
		return genome(name = self.name,
			genes = genes,
			sequences = self.sequences,
			chromosomes = self.chromosomes | chromosomes,
			annotation = annotation,
			annotationPath = path,
			seqPath = self.seqPath,
			seqLens = self.seqLens)
	
	def setSequences(self, seqs):
		return genome(name = self.name,
			genes = self.genes,
			sequences = seqs,
			chromosomes = self.chromosomes | set(s.name for s in seqs),
			annotation = self.annotation,
			annotationPath = self.annotationPath,
			seqPath = str(seqs),
			seqLens = None)
	
	def streamFASTA(self, path):
		return self.setSequences(streamFASTA(path))
	
	def streamFASTAGZ(self, path):
		return self.setSequences(streamFASTAGZ(path))
	
	def stream2bit(self, path):
		return self.setSequences(stream2bit(path))
	
	def sequenceLengths(self):
		if self.sequences is None:
			return {}
		if self.seqLens is None:
			self.seqLens = self.sequences.sequenceLengths()
		return self.seqLens
	
	def getCDS(self):
		return regions('CDS', [ r for g in self.genes for r in g.CDS ])
	
	def gene(self, name):
		return self.genesByName[name]
	
	def windows(self, size, step):
		return self.sequences.windows(size = size, step = step)
	
	def windowRegions(self, size, step):
		return self.sequences.windowRegions(size = size, step = step)
	
	def __str__(self):
		return 'Genome<%s; annotation: %s; sequences: %s>'%(self.name,
			'' if self.annotationPath is None else self.annotationPath,
			'' if self.seqPath is None else self.seqPath)
	
	def _as_dict_(self):
		return {
			'Seq.': [ g.region.seq for g in self.genes ],
			'Start': [ g.region.start for g in self.genes ],
			'End': [ g.region.end for g in self.genes ],
			'Strand': [ '+' if g.region.strand else '-' for g in self.genes ],
			'Length': [ len(g.region) for g in self.genes ],
			'Name': [ g.name for g in self.genes ],
			'Synonyms': [ '; '.join(g.synonyms) for g in self.genes ],
			'Exons': [ len(g.exons) for g in self.genes ],
			'CDS': [ len(g.CDS) for g in self.genes ],
		}
	
	def table(self):
		return nctable(
			'Genome: ' + self.__str__(),
			self._as_dict_(),
			align = { 'Seq.': 'l', 'Name': 'l', 'Synonyms': 'l' },
			#strcroplen = 0
		)
	
	def _repr_html_(self):
		return self.table()._repr_html_()
	
	def __repr__(self):
		return self.table().__repr__()


############################################################################
# Genome plotting

def drawRoundBox(ax, start, end, width, y, height, fc, ec, maxYB, rlabel = None, clip_on = True):
	import matplotlib.pyplot as plt
	from matplotlib.patches import FancyBboxPatch
	w = end - start + 1
	rs = min(w * 0.5 * width * 0.00001, 500. * width * 0.00001)
	patch = FancyBboxPatch((start, y),
		w, height,
		boxstyle="round,rounding_size=" + str(rs),
		mutation_aspect = 2.*maxYB/width,
		fc = fc,
		ec = ec,
		clip_on = clip_on)
	ax.add_patch(patch)
	if rlabel is not None:
		plt.text(end + width*0.015,
					y + height*.5,
					rlabel,
					verticalalignment = 'center',
					horizontalalignment = 'left',
					color = ax.yaxis.label.get_color(),
					fontsize = 10,
					clip_on = False)
	return patch

def drawBox(ax, start, end, width, y, height, fc, ec, maxYB, rlabel = None, clip_on = True):
	import matplotlib.pyplot as plt
	from matplotlib.patches import FancyBboxPatch
	w = end - start + 1
	rs = min(w * width * 0.00001, 500. * width * 0.00001)
	patch = FancyBboxPatch((start, y),
		w, height,
		boxstyle="square",
		mutation_aspect = 2.*maxYB/width,
		fc = fc,
		ec = ec,
		clip_on = clip_on)
	ax.add_patch(patch)
	if rlabel is not None:
		plt.text(end + width*0.015,
					y + height*.5,
					rlabel,
					verticalalignment = 'center',
					horizontalalignment = 'left',
					color = ax.yaxis.label.get_color(),
					fontsize = 10,
					clip_on = False)
	return patch

def drawGeneBody(ax, start, end, width, y, height, maxYB, rlabel = None, clip_on = True):
	colBody = [ 0.75, 0.75, 0.75 ]
	colInvisible = [ 0., 0., 0., 0. ]
	return drawRoundBox(ax = ax, start = start, end = end, width = width,
		y = y, height = height, maxYB = maxYB,
		fc = colBody + [0.4],
		ec = colInvisible,
		rlabel = rlabel,
		clip_on = clip_on)

def drawGeneOutline(ax, start, end, width, y, height, maxYB, clip_on = True):
	colBorder = [ 0.1, 0.1, 0.1, 1. ]
	colInvisible = [ 0., 0., 0., 0. ]
	drawRoundBox(ax = ax, start = start, end = end, width = width,
		y = y, height = height, maxYB = maxYB,
		fc = colInvisible,
		ec = colBorder,
		clip_on = clip_on)

def drawGeneExon(ax, start, end, y, width, height, maxYB, rlabel = None, clip_on = True):
	colExons = [ 0.45, 0.45, 0.45 ]
	colInvisible = [ 0., 0., 0., 0. ]
	drawBox(ax = ax, start = start, end = end, width = width,
		y = y, height = height, maxYB = maxYB,
		fc = colExons + [1.],
		ec = colInvisible,
		rlabel = rlabel,
		clip_on = clip_on)

def drawGeneCDS(ax, start, end, y, width, height, maxYB, rlabel = None, clip_on = True):
	colCDS = [ 0.1, 0.1, 0.1 ]
	colInvisible = [ 0., 0., 0., 0. ]
	drawBox(ax = ax, start = start, end = end, width = width,
		y = y, height = height, maxYB = maxYB,
		fc = colCDS + [1.],
		ec = colInvisible,
		rlabel = rlabel,
		clip_on = clip_on)

#------------------------------------

class GenomePlotTrack:
	def __init__(self):
		pass
	def getHeight(self):
		return 0.0
	def getName(self):
		return ''
	def draw(self, ax, chromosome, coordStart, coordEnd, width, rmargin, maxYB, colour):
		pass

class GenomePlotTrackRegions(GenomePlotTrack):
	def __init__(self, rs):
		self.rs = rs
	def getHeight(self):
		return 1.0
	def getName(self):
		return self.rs.name
	def draw(self, ax, chromosome, coordStart, coordEnd, width, rmargin, maxYB, colour):
		frs = self.rs.filter('', lambda r: r.seq == chromosome\
			and r.end >= coordStart\
			and r.start <= coordEnd)
		height = self.yB - self.yA
		for r in frs:
			drawRoundBox(ax = ax, start = r.start, end = r.end, width = width,
				y = self.yA + rmargin, height = height - 2*rmargin, maxYB = maxYB,
				fc = colour[:3] + [0.4],
				ec = [ c*0.5 for c in colour[:3] ] + [1.])

class GenomePlotTrackGenome(GenomePlotTrack):
	def __init__(self, rs):
		self.rs = rs
	def getHeight(self):
		return 3.0
	def getName(self):
		return self.rs.name
	def draw(self, ax, chromosome, coordStart, coordEnd, width, rmargin, maxYB, colour):
		height = self.yB - self.yA
		cgenes = [
			g for g in self.rs.genes
			if g.region.seq == chromosome\
				 and g.region.end >= coordStart\
				 and g.region.start <= coordEnd
		]
		cy = height / 2.
		geneheight = 1. - 2.*rmargin
		colInvisible = [ 0., 0., 0., 0. ]
		# Legend: Gene body
		drawGeneBody(ax = ax, start = coordEnd + width*0.01, end = coordEnd + width*0.01 + 3. * 0.006 * width, width = width, y = self.yB - 0.1 - geneheight, height = geneheight, maxYB = maxYB, rlabel = 'Gene body', clip_on = False)
		drawGeneOutline(ax = ax, start = coordEnd + width*0.01, end = coordEnd + width*0.01 + 3. * 0.006 * width, width = width, y = self.yB - 0.1 - geneheight, height = geneheight, maxYB = maxYB, clip_on = False)
		# Legend: Gene exon
		drawGeneExon(ax = ax, start = coordEnd + width*0.01, end = coordEnd + width*0.01 + 3. * 0.006 * width, width = width, y = self.yB - 0.1 - geneheight - 0.1 - geneheight, height = geneheight, maxYB = maxYB, rlabel = 'Exon', clip_on = False)
		# Legend: Gene CDS
		drawGeneCDS(ax = ax, start = coordEnd + width*0.01, end = coordEnd + width*0.01 + 3. * 0.006 * width, width = width, y = self.yB - 0.1 - geneheight - 0.1 - geneheight - 0.1 - geneheight, height = geneheight, maxYB = maxYB, rlabel = 'CDS', clip_on = False)
		#
		for strand in [ False, True ]:
			gh = 0.02
			exonh = 0.2
			ccy = self.yA + (cy + 0.5 if strand else cy - 0.5)
			for g in cgenes:
				if g.region.strand != strand: continue
				# Gene body
				drawGeneBody(ax = ax, start = g.region.start, end = g.region.end, width = width, y = ccy - geneheight*.5, height = geneheight, maxYB = maxYB)
				# Exons
				for r in g.exons:
					drawGeneExon(ax = ax, start = r.start, end = r.end, width = width, y = ccy - geneheight*.5, height = geneheight, maxYB = maxYB)
				# CDS
				for r in g.CDS:
					drawGeneCDS(ax = ax, start = r.start, end = r.end, width = width, y = ccy - geneheight*.5, height = geneheight, maxYB = maxYB)
				#
				drawGeneOutline(ax = ax, start = g.region.start, width = width, end = g.region.end, y = ccy - geneheight*.5, height = geneheight, maxYB = maxYB)
			for g in cgenes:
				if g.region.strand == strand:
					center = (g.region.start + g.region.end)/2.
					if center <= coordStart or center >= coordEnd: continue
					ax.text(center,
						ccy + (0.65 if strand else -0.65),
						g.name,
						horizontalalignment = 'center',
						verticalalignment = 'center')

class GenomePlotTrackCurve(GenomePlotTrack):
	def __init__(self, rs):
		self.rs = rs
	def getHeight(self):
		if self.rs.thresholdValue is not None:
			return 3.0
		else:
			return 1.0
	def getName(self):
		return self.rs.name
	def draw(self, ax, chromosome, coordStart, coordEnd, width, rmargin, maxYB, colour):
		import matplotlib.pyplot as plt
		from matplotlib.patches import Rectangle, Polygon
		from matplotlib.lines import Line2D
		colInvisible = [ 0., 0., 0., 0. ]
		height = self.yB - self.yA
		ccurve = next((c for c in self.rs if c.seq == chromosome), None)
		if ccurve == None: return
		# Rasterize - logic required depends on curve type
		if isinstance(ccurve, fixedStepCurve):
			iA = max(int((coordStart - ccurve.span) / ccurve.step), 0)
			iB = max(
					min(int((coordEnd) / ccurve.step), len(ccurve)-1),
				0)
			span = ccurve.span
			step = ccurve.step
			vals = ccurve[iA:iB+1]
			#
			res = 2048
			srcXA = iA * ccurve.step #coordStart
			srcXB = iB * ccurve.step #coordEnd
			scale = res / (srcXB - srcXA)
			ret = [ None for _ in range(res) ]
			for i, v in enumerate(vals):
				dstXA = int(min(max(
					i * ccurve.step * res / ((iB - iA) * ccurve.step),
					0),
					res-1))
				dstXB = int(min(max(
					res * (i + ccurve.span / ccurve.step) / (iB - iA),
					0),
					res-1))
				for x in range(dstXA, dstXB+1):
					Q = (ccurve.step * (iA + (iB - iA) * float(x) / res), v)
					if ret[x] is None or Q[1] > ret[x][1]:
						ret[x] = Q
			raster = [ v for v in ret if v is not None ]
		elif isinstance(ccurve, variableStepCurve):
			vals = [
				v for v in ccurve
				if v.start+v.span >= coordStart
				and v.start <= coordEnd
			]
			#
			res = 2048
			srcXA = coordStart
			srcXB = coordEnd
			scale = res / (srcXB - srcXA)
			ret = [ None for _ in range(res) ]
			for v in vals:
				dstXA = int(min(max((v.start - srcXA) * scale, 0), res-1))
				dstXB = int(min(max((v.start + v.span - srcXA) * scale, 0), res-1))
				for x in range(dstXA, dstXB+1):
					ret[x] = (srcXA + (float(x) / scale), v.value)
			raster = [ v for v in ret if v is not None ]
		# Plot curve
		vBase = 0.
		vMin = min(y for x, y in raster)
		vMax = max(y for x, y in raster)
		yBottom = min(vBase, vMin)
		yTop = vMax
		cheight = height
		predHeight = 1.
		def V2Y(v):
			return self.yA + rmargin + ((v + vBase - yBottom) * vScale)
		if self.rs.thresholdValue is not None:
			yBottom = min(yBottom, self.rs.thresholdValue)
			yTop = max(yTop, self.rs.thresholdValue)
			cheight -= predHeight
		vScale = (cheight - rmargin*2.) / (yTop - yBottom)
		polypts = [ ]
		polypts += [
			(x, V2Y(max(y, vBase)))
			for x, y in raster ]
		polypts += reversed([
			(x, V2Y(min(y, vBase)))
			for x, y in raster ])
		poly = Polygon(
			polypts,
			fc = colour[:3] + [0.4],
			ec = colInvisible)
		ax.add_patch(poly)
		ax.add_line(Line2D([ x for x, y in raster ],
			[ V2Y(y) for x, y in raster ],
			linewidth = 0.8,
			color = [ c*0.75 for c in colour[:3] ] + [0.75]))
		# Legend
		lyA = V2Y(vMin)
		lyB = V2Y(vMax)
		ax.add_patch(Rectangle(
			(coordEnd, lyA),
			1. * 0.006 * width,
			lyB-lyA,
			fc = (0.2, 0.2, 0.2, 1.),
			ec = colInvisible,
			clip_on = False
		))
		tticks = [ (vMax, str(vMax)), (vMin, str(vMin)) ]
		if self.rs.thresholdValue is not None:
			tticks.append((self.rs.thresholdValue, str(self.rs.thresholdValue) + ' (threshold)'))
		if vMin < 0. and vMax > 0.:
			tticks.append((0.0, str(0.0)))
		tticks = sorted(tticks, key = lambda p: p[0])
		for v, n in tticks:
			ly = V2Y(v)
			ax.add_line(Line2D([ coordEnd + 0. * 0.006 * width, coordEnd + 2. * 0.006 * width ],
				[ ly, ly ],
				color = (0.2, 0.2, 0.2, 1.),
				linewidth = 1.,
				clip_on = False))
			plt.text(coordEnd + 3. * 0.006 * width,
					ly,
					n,
					verticalalignment = 'center',
					horizontalalignment = 'left',
					color = ax.yaxis.label.get_color(),
					fontsize = 8,
					clip_on = False)
		# Thresholded
		if self.rs.thresholdValue is not None:
			tthr = V2Y(self.rs.thresholdValue)
			ax.plot(
				[ coordStart, coordEnd ],
				[ tthr, tthr ],
				linestyle = '-',
				color = 'grey',
				label = 'Expected at random')
			frs = self.rs.regions().filter('', lambda r: r.seq == chromosome\
				and r.end >= coordStart\
				and r.start <= coordEnd)
			for r in frs:
				drawRoundBox(ax = ax, start = r.start, end = r.end, width = width,
					y = self.yB - predHeight + rmargin, height = predHeight - 2*rmargin,
					maxYB = maxYB,
					fc = colour[:3] + [0.4],
					ec = [ c*0.5 for c in colour[:3] ] + [1.])
			# Legend
			geneheight = 1. - 2.*rmargin
			drawRoundBox(ax = ax, start = coordEnd + width*0.01, width = width,
				end = coordEnd + width*0.01 + 3. * 0.006 * width,
				y = self.yB - 0.1 - geneheight, maxYB = maxYB,
				height = geneheight,
				fc = colour[:3] + [0.4],
				ec = [ c*0.5 for c in colour[:3] ] + [1.],
				rlabel = 'Thresholded',
				clip_on = False)

class GenomePlotTrackCVPredictions:
	def __init__(self, rs):
		self.rs = rs
	def getHeight(self):
		# TODO Add multiclass support
		bpred = self.rs.predictionCurves[positive][0]
		if bpred.thresholdValue is not None:
			return 3.0 + (1. if self.rs.hasCores else 0.)
		else:
			return 1.0
	def getName(self):
		return self.rs.name
	def draw(self, ax, chromosome, coordStart, coordEnd, width, rmargin, maxYB, colour):
		import matplotlib.pyplot as plt
		from matplotlib.patches import Rectangle, Polygon
		from matplotlib.lines import Line2D
		colInvisible = [ 0., 0., 0., 0. ]
		height = self.yB - self.yA
		# TODO Add multiclass support
		bpreds = self.rs.predictionCurves[positive]
		bpred = bpreds[0]
		ccurves = [
			next((c for c in p if c.seq == chromosome), None)
			for p in bpreds
		]
		ccurve = ccurves[0]
		if ccurve == None: return
		# Rasterize - logic required depends on curve type
		if not isinstance(ccurve, fixedStepCurve):
			raise Exception('Error: Prediction curve should be of type fixedStepCurve')
		iA = max(int((coordStart - ccurve.span) / ccurve.step), 0)
		iB = max(
				min(int((coordEnd) / ccurve.step), len(ccurve)-1),
			0)
		span = ccurve.span
		step = ccurve.step
		#
		res = 2048
		srcXA = iA * ccurve.step #coordStart
		srcXB = iB * ccurve.step #coordEnd
		scale = res / (srcXB - srcXA)
		ret = [ None for _ in range(res) ]
		for i in range(1+iB-iA):
			vals = [ c[i+iA] for c in ccurves ]
			vMean = mean(vals)
			vCI = CI(vals)
			dstXA = int(min(max(
				i * ccurve.step * res / ((iB - iA) * ccurve.step),
				0),
				res-1))
			dstXB = int(min(max(
				res * (i + ccurve.span / ccurve.step) / (iB - iA),
				0),
				res-1))
			for x in range(dstXA, dstXB+1):
				Q = (ccurve.step * (iA + (iB - iA) * float(x) / res), vMean, vCI)
				if ret[x] is None or Q[1] > ret[x][1]:
					ret[x] = Q
		raster = [ v for v in ret if v is not None ]
		# Plot curve
		vBase = 0.
		vMin = min(y for x, y, ci in raster)
		vMax = max(y for x, y, ci in raster)
		yBottom = min(vBase, vMin)
		yTop = vMax
		cheight = height
		predHeight = 1.
		def V2Y(v):
			return self.yA + rmargin + ((v + vBase - yBottom) * vScale)
		if bpred.thresholdValue is not None:
			yBottom = min(yBottom, bpred.thresholdValue)
			yTop = max(yTop, bpred.thresholdValue)
			cheight -= predHeight
			if self.rs.hasCores:
				cheight -= predHeight
		vScale = (cheight - rmargin*2.) / (yTop - yBottom)
		
		polypts = [ ]
		polypts += [
			(x, V2Y(y+ci))
			for x, y, ci in raster ]
		polypts += reversed([
			(x, V2Y(y-ci))
			for x, y, ci in raster ])
		poly = Polygon(
			polypts,
			fc = colour[:3] + [0.3],
			ec = colInvisible)
		ax.add_patch(poly)
		
		polypts = [ ]
		polypts += [
			(x, V2Y(max(y, vBase)))
			for x, y, ci in raster ]
		polypts += reversed([
			(x, V2Y(min(y, vBase)))
			for x, y, ci in raster ])
		poly = Polygon(
			polypts,
			fc = colour[:3] + [0.5],
			ec = colInvisible)
		
		ax.add_patch(poly)
		ax.add_line(Line2D([ x for x, y, ci in raster ],
			[ V2Y(y) for x, y, ci in raster ],
			linewidth = 0.8,
			color = [ c*0.75 for c in colour[:3] ] + [0.75]))
		# Legend
		lyA = V2Y(vMin)
		lyB = V2Y(vMax)
		ax.add_patch(Rectangle(
			(coordEnd, lyA),
			1. * 0.006 * width,
			lyB-lyA,
			fc = (0.2, 0.2, 0.2, 1.),
			ec = colInvisible,
			clip_on = False
		))
		tticks = [ (vMax, str(vMax)), (vMin, str(vMin)) ]
		thrMean = None
		if bpred.thresholdValue is not None:
			thr = [ p.thresholdValue for p in bpreds ]
			thrMean = mean(thr)
			thrCI = CI(thr)
			tticks.append((thrMean, '%.2f +/- %.2f'%(thrMean, thrCI) + ' (threshold)'))
		if vMin < 0. and vMax > 0.:
			tticks.append((0.0, str(0.0)))
		tticks = sorted(tticks, key = lambda p: p[0])
		for v, n in tticks:
			ly = V2Y(v)
			ax.add_line(Line2D([ coordEnd + 0. * 0.006 * width, coordEnd + 2. * 0.006 * width ],
				[ ly, ly ],
				color = (0.2, 0.2, 0.2, 1.),
				linewidth = 1.,
				clip_on = False))
			plt.text(coordEnd + 3. * 0.006 * width,
					ly,
					n,
					verticalalignment = 'center',
					horizontalalignment = 'left',
					color = ax.yaxis.label.get_color(),
					fontsize = 8,
					clip_on = False)
		# Thresholded
		if bpred.thresholdValue is not None:
			tthr = V2Y(thrMean)
			tthrCI = V2Y(thrMean + thrCI) - tthr
			ax.add_patch(Rectangle(
				(coordStart, tthr - tthrCI),
				coordEnd - coordStart,
				tthrCI * 2.,
				fc = (0.4, 0.4, 0.4, 0.3),
				ec = colInvisible,
				clip_on = False
			))
			ax.plot(
				[ coordStart, coordEnd ],
				[ tthr, tthr ],
				linestyle = '-',
				color = 'grey',
				label = 'Expected at random')
			for rs in self.rs.regions(label = positive):
				frs = rs.filter('', lambda r: r.seq == chromosome\
					and r.end >= coordStart\
					and r.start <= coordEnd)
				for r in frs:
					drawRoundBox(ax = ax, start = r.start, end = r.end, width = width,
						y = self.yB - predHeight + rmargin, height = predHeight - 2*rmargin,
						maxYB = maxYB,
						fc = colour[:3] + [0.4 / len(bpreds)],
						ec = [ c*0.5 for c in colour[:3] ] + [1. / len(bpreds)])
			if self.rs.hasCores:
				for c in self.rs.predictionCurves[positive]:
					frs = c.regions().filter('', lambda r: r.seq == chromosome\
						and r.end >= coordStart\
						and r.start <= coordEnd)
					for r in frs:
						drawRoundBox(ax = ax, start = r.start, end = r.end, width = width,
							y = self.yB - predHeight*2 + rmargin, height = predHeight - 2*rmargin,
							maxYB = maxYB,
							fc = colour[:3] + [0.4 / len(bpreds)],
							ec = [ c*0.5 for c in colour[:3] ] + [1. / len(bpreds)])
			# Legend
			geneheight = 1. - 2.*rmargin
			drawRoundBox(ax = ax, start = coordEnd + width*0.01, width = width,
				end = coordEnd + width*0.01 + 3. * 0.006 * width,
				y = self.yB - 0.1 - geneheight, maxYB = maxYB,
				height = geneheight,
				fc = colour[:3] + [0.4],
				ec = [ c*0.5 for c in colour[:3] ] + [1.],
				rlabel = 'Cores' if self.rs.hasCores else 'Thresholded',
				clip_on = False)
			if self.rs.hasCores:
				drawRoundBox(ax = ax, start = coordEnd + width*0.01, width = width,
					end = coordEnd + width*0.01 + 3. * 0.006 * width,
					y = self.yB - 0.1*2 - geneheight*2, maxYB = maxYB,
					height = geneheight,
					fc = colour[:3] + [0.4],
					ec = [ c*0.5 for c in colour[:3] ] + [1.],
					rlabel = 'Thresholded',
					clip_on = False)

#------------------------------------

def plotGenomeTracks(tracks, chromosome, coordStart, coordEnd, style = 'ggplot', outpath = None, figsize = None):
	if figsize is None:
		figsize = (13., 1. * len(tracks))
	def coordFmt(t):
		if t >= 1000000:
			return '%.3f Mb'%(t / 1000000.)
		if t >= 1000:
			return '%.3f kb'%(t / 1000.)
		return '%.3f bp'%t
	try:
		import matplotlib.pyplot as plt
		import base64
		from io import BytesIO
		from IPython.core.display import display, HTML
		from matplotlib.transforms import ScaledTranslation
		import matplotlib.ticker as mticker
		import struct
		with plt.style.context(style):
			fig, ax = plt.subplots(1, figsize = figsize)
			# Set up axes
			plt.yticks(range(len(tracks)))
			ax.set_xlim(coordStart, coordEnd)
			ax.set_ylim(0, len(tracks))
			ticks = ax.get_xticks().tolist()
			ax.xaxis.set_major_locator(mticker.FixedLocator(ticks))
			ax.set_xticklabels([ coordFmt(t) for t in ticks ], fontsize = 12)
			ax.set_yticklabels([  ])
			setYA = []
			setYB = []
			width = coordEnd - coordStart
			# Create plotting tracks
			y = 0.
			ptracks = []
			for rs in tracks:
				yA = y
				nrs = None
				if isinstance(rs, GenomePlotTrack):
					nrs = rs
				elif isinstance(rs, genome):
					nrs = GenomePlotTrackGenome(rs)
				elif isinstance(rs, curves):
					nrs = GenomePlotTrackCurve(rs)
				elif isinstance(rs, CVModelPredictions):
					nrs = GenomePlotTrackCVPredictions(rs)
				else:
					nrs = GenomePlotTrackRegions(rs)
				y += nrs.getHeight()
				ptracks.append(nrs)
				nrs.yA = yA
				nrs.yB = y
				setYA.append(yA)
				setYB.append(y)
			# Plot names
			for rs in ptracks:
				plt.text(coordStart - width*0.015,
					(rs.yB + rs.yA) / 2. + 0.05,
					rs.getName(),
					verticalalignment = 'top',
					horizontalalignment = 'right',
					color = ax.yaxis.label.get_color(),
					fontsize = 15,
					rotation = 45,
					clip_on = False)
			# Plot ticks and labels
			ax.set_yticks([ 0. ] + setYB)
			ax.yaxis.set_label_coords(-0.3, 1.5)
			plt.setp(ax.get_yticklabels(), rotation=45, ha="right",
				rotation_mode="anchor")
			plt.tick_params(
				axis='y',
				which='both',
				left=False,
				right=False)
			# Shift y axis labels
			dx = -0.1
			dy = 0.2
			offset = ScaledTranslation(dx, dy, fig.dpi_scale_trans)
			for label in ax.yaxis.get_majorticklabels():
				label.set_transform(label.get_transform() + offset)
			# Draw
			rmargin = 0.2
			for rsi, rs in enumerate(ptracks):
				colourRaw = plt.rcParams['axes.prop_cycle'].by_key()['color'][rsi % len(plt.rcParams['axes.prop_cycle'].by_key()['color'])]
				colour = [ v/255. for v in struct.unpack('BBB', bytes.fromhex(colourRaw[1:])) ]
				rs.draw(ax = ax, chromosome = chromosome, coordStart = coordStart, coordEnd = coordEnd, width = width, rmargin = rmargin,
					maxYB = max(setYB), colour = colour)
			#
			plt.xlabel('Chromosome ' + chromosome, fontsize=18)
			fig.tight_layout()
			if outpath is None:
				bio = BytesIO()
				fig.savefig(bio, format='png')
				plt.close('all')
				encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
				html = '<img src=\'data:image/png;base64,%s\'>'%encoded
				display(HTML(html))
			else:
				fig.savefig(outpath)
				plt.close('all')
	#
	except ImportError as err:
		raise err




#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################
# Interfacing with TensorFlow

from .sequences import sequences, positive, negative, sequenceStream
from .models import sequenceModel
import numpy as np
import tensorflow as tf
from tensorflow import keras

def setSeed(seed):
	tf.random.set_seed(seed)

ntmat = {
	'A': [ 1., 0., 0., 0. ],
	'T': [ 0., 1., 0., 0. ],
	'G': [ 0., 0., 1., 0. ],
	'C': [ 0., 0., 0., 1. ],
	'N': [ 0., 0., 0., 0. ],
}

def getSequenceMatrix(seq):
	return np.array([ [ ntmat[nt] ] for nt in seq ])

class KerasModel(sequenceModel):
	def __init__(self, name, windowSize, windowStep, modelConstructor, epochs = 100, targetLabel = positive, labels = [ positive, negative ], trainingSet = None, batchsize = 1000, trainWindows = False):
		super().__init__(name, enableMultiprocessing = False)
		self.modelConstructor = modelConstructor
		self.windowSize, self.windowStep = windowSize, windowStep
		self.targetLabel, self.labels = targetLabel, list(set([targetLabel]) | set(labels))
		self.nLabels = len(self.labels)
		self.threshold = 0.0
		self.epochs = epochs
		self.cls = None
		self.batchsize = batchsize
		self.trainingSet = trainingSet
		self.trainWindows = trainWindows
		if trainingSet is not None:
			self._train(trainingSet)

	def _train(self, trainingSet):
		model = self.modelConstructor()
		
		#----------------------------------------------------------

		model.compile(
			optimizer = 'adam',
			loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
			metrics = ['accuracy']
		)

		self.labelValues = {
			label: i
			for i, label in enumerate(self.labels)
		}
		if self.trainWindows:
			train_seq = np.array([
				getSequenceMatrix(win.seq)
				for label in self.labels
				for seq in trainingSet.withLabel(label)
				for win in seq.windows(self.windowSize, self.windowStep)
			])
			train_labels = np.array([
				self.labelValues[label]
				for label in self.labels
				for seq in trainingSet.withLabel(label)
				for win in seq.windows(self.windowSize, self.windowStep)
			])
		else:
			train_seq = np.array([
				getSequenceMatrix(seq.seq)
				for label in self.labels
				for seq in trainingSet.withLabel(label)
			])
			train_labels = np.array([
				self.labelValues[label]
				for label in self.labels
				for seq in trainingSet.withLabel(label)
			])

		model.fit(
			train_seq,
			train_labels,
			epochs = self.epochs,
			verbose = 0
		)

		self.cls = model

	def getSequenceScores(self, seqs, nStreamFetch = 100000, maxStreamFetchNT = 100000000):
		if self.batchsize == 0:
			return super().getSequenceScores(seqs)
		if isinstance(seqs, sequenceStream):
			def fetchseq():
				i = 0
				for blk in seqs.fetch(nStreamFetch, maxStreamFetchNT):
					for cseq in blk:
						for win in cseq.windows(self.windowSize, self.windowStep):
							if len(win) != self.windowSize:
								continue
							#
							yield (i, win)
						i += 1
			#
			seqwinit = (
				(i, win)
				for i, win in fetchseq()
			)
		else:
			seqwinit = (
				(i, win)
				for i, cseq in enumerate(seqs)
				for win in cseq.windows(self.windowSize, self.windowStep)
				if len(win) == self.windowSize
			)
		seqscores = [  ]
		while True:
			batch = [
				b for b in [
					next(seqwinit, None) for _ in range(self.batchsize)
				]
				if b is not None
			]
			if len(batch) == 0: break
			p = self.cls.predict(np.array([getSequenceMatrix(seq.seq) for i, seq in batch]))
			i = self.labelValues[self.targetLabel]
			scores = p[:, i]
			#scores = p[:, i] * p[:, i] / (sum(p[:, n] for n in range(self.nLabels)) + 0.1) # Pseudocount to avoid zero division
			for score, (i, win) in zip(scores, batch):
				if i >= len(seqscores):
					seqscores += [ -float('INF') for _ in range(i - len(seqscores) + 1) ]
				if score > seqscores[i]:
					seqscores[i] = score
		#
		return seqscores

	def scoreWindow(self, seq):
		p = self.cls.predict(np.array([ getSequenceMatrix(seq.seq) ]))
		i = self.labelValues[self.targetLabel]
		return p[0, i]
		#return p[0, i] * p[0, i] / (sum(p[0, n] for n in range(self.nLabels)) + 0.1)

	def getTrainer(self):
		return lambda ts: KerasModel(self.name, trainingSet = ts, windowSize = self.windowSize, windowStep = self.windowStep, epochs = self.epochs, targetLabel = self.targetLabel, labels = self.labels, batchsize = self.batchsize, modelConstructor = self.modelConstructor)

	def __str__(self):
		return 'Multi-class Keras model <Training set: %s; Labels: %s; Target label: %s; Epochs: %d>'%(str(self.trainingSet), str(self.labels), str(self.targetLabel), self.epochs)

	def __repr__(self): return self.__str__()

class DeepMOCCA(KerasModel):
	def __init__(self, name, windowSize, windowStep, nConv = 20, convLen = 10, epochs = 100,
				 targetLabel = positive, labels = [ positive, negative ], trainingSet = None,
				 batchsize = 1000, trainWindows = True,
				 includeDinucleotides = True,
				 pairingDistance = 500):
		self.nConv, self.convLen = nConv, convLen
		self.includeDinucleotides = includeDinucleotides
		self.pairingDistance = pairingDistance
		modelConstructor = lambda: self.constructModel(windowSize = windowSize,
						nConv = nConv, convLen = convLen, labels = labels)
		super().__init__(name = name, windowSize = windowSize, windowStep = windowStep,
						modelConstructor = modelConstructor, epochs = epochs,
						targetLabel = targetLabel, labels = labels,
						trainingSet = trainingSet,
						batchsize = batchsize, trainWindows = trainWindows)

	def constructModel(self, windowSize, nConv, convLen, labels):
		scaleFac = 10
		bloat = int((min(self.pairingDistance, windowSize)-1)/scaleFac)
		hbloat = int(bloat/2)
		nLabels = len(labels)

		#----------------------------------------------------------
		# Model structure
		I = keras.Input(shape = (None, 1, 4))

		# - Merges down NT channels and outputs nConv convolutions
		C1 = keras.layers.Conv2D(nConv, kernel_size = (convLen, 1), activation = 'relu',
									  data_format = "channels_last")(I)
		
		if self.includeDinucleotides:
			C1P = keras.layers.ZeroPadding2D(padding = (int(convLen / 2) - 1, 0))(C1)

			# Dinucleotide convolutions
			# - Merges down NT channels and outputs nConv convolutions
			C2 = keras.layers.Conv2D(nConv, kernel_size = (2, 1), activation = 'relu',
										  data_format = "channels_last")(I)
			C2P = C2

			Combine = keras.layers.Concatenate(axis=-1)([ C1P, C2P ])
		else:
			Combine = C1

		# Reduces resolution (but not channels/convolutions)
		DS = keras.layers.AveragePooling2D((scaleFac, 1))(Combine)

		# Bloating - Averages convolution outputs over a window
		BLT0 = keras.layers.Permute((1, 3, 2))(DS)
		BLT1 = keras.layers.Conv2D(1, kernel_size = (bloat, 1),
									  activation=None,
									  trainable = False,
									  use_bias=False,
									  weights = [
										tf.constant(
										   np.array([ 1./bloat for _ in range(bloat) ])
											 .reshape((bloat, 1, 1, 1))),
									  ],
									  data_format = "channels_last")(BLT0)
		BLT = keras.layers.Permute((1, 3, 2))(BLT1)

		# Merges convolutions into mixing convolutions
		C2 = keras.layers.Conv2D(nConv, kernel_size = (1, 1), activation = 'relu',
									  data_format = "channels_last")(BLT)
		Pool = keras.layers.GlobalMaxPooling2D(data_format = "channels_last")(C2)

		# Max pooling of mixing convolutions
		Term = keras.layers.Dense(nLabels, activation=tf.nn.softmax)(Pool)

		model = keras.Model(I, Term)
		
		return model
		
	def getTrainer(self):
		return lambda ts: DeepMOCCA(self.name,
			trainingSet = ts, windowSize = self.windowSize,
			windowStep = self.windowStep, nConv = self.nConv,
			convLen = self.convLen, epochs = self.epochs,
			targetLabel = self.targetLabel, labels = self.labels,
			batchsize = self.batchsize, trainWindows = self.trainWindows,
			includeDinucleotides = self.includeDinucleotides,
			pairingDistance = self.pairingDistance)

	def __str__(self):
		return 'Deep-MOCCA <Training set: %s; Labels: %s; Target label: %s; Convolutions: %d; Convolution length: %d; Pairing distance: %d; Include dinucleotides: %s; Epochs: %d>'%(str(self.trainingSet), str(self.labels), str(self.targetLabel), self.nConv, self.convLen, self.pairingDistance, 'Yes' if self.pairingDistance else 'No', self.epochs)

	def __repr__(self): return self.__str__()




#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################
# Interfacing with scikit-learn

from .features import featureScaler
from .sequences import sequences, positive, negative
from .models import sequenceModel
import numpy as np
from sklearn import svm
from sklearn.ensemble import RandomForestClassifier
import random
from sklearn.linear_model import Lasso
from .featurenetwork import baseModel

# Support Vector Machines - Sequence model
class sequenceModelSVM(sequenceModel):
	"""
	The `sequenceModelSVM` class trains a Support Vector Machine (SVM) using scikit-learn.
	
	:param name: Model name.
	:param features: Feature set.
	:param positives: Positive training sequences.
	:param negatives: Negative training sequences.
	:param windowSize: Window size.
	:param windowStep: Window step size.
	:param kDegree: Kernel degree.
	
	:type name: str
	:type features: features
	:type positives: sequences
	:type negatives: sequences
	:type windowSize: int
	:type windowStep: int
	:type kDegree: float
	"""
	
	def __init__(self, name, features, trainingSet, windowSize, windowStep, kDegree, scale = True, labelPositive = positive, labelNegative = negative):
		super().__init__(name)
		self.windowSize, self.windowStep = windowSize, windowStep
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.trainingSet = trainingSet
		positives, negatives = trainingSet.withLabel([ labelPositive, labelNegative ])
		wPos = sequences(positives.name, [ w for s in positives for w in s.windows(self.windowSize, self.windowStep) ])
		wNeg = sequences(negatives.name, [ w for s in negatives for w in s.windows(self.windowSize, self.windowStep) ])
		if scale:
			features = featureScaler( features, trainingSet = wPos + wNeg )
		self.scale = scale
		self.features = features
		self.kernel = kDegree
		self.threshold = 0.0
		vP = [ self.getSequenceFeatureVector(w) for w in wPos ]
		vN = [ self.getSequenceFeatureVector(w) for w in wNeg ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		self.cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		self.cls.fit( np.array(vP+vN), np.array(cP+cN) )
	
	def getTrainer(self):
		return lambda ts: sequenceModelSVM(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, kDegree = self.kernel, scale = self.scale, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
	
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	
	def scoreWindow(self, seq):
		return self.cls.decision_function(np.array([self.getSequenceFeatureVector(seq)]))[0]
	
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Training set: %s; Positive label: %s; Negative label: %s; Kernel: %s; Support vectors: %d>'%(str(self.features), len(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], len(self.cls.support_vectors_))
	
	def __repr__(self): return self.__str__()

# Random Forest - Sequence model
class sequenceModelRF(sequenceModel):
	"""
	The `sequenceModelRF` class trains a Random Forest (RF) model using scikit-learn.
	
	:param name: Model name.
	:param features: Feature set.
	:param positives: Positive training sequences.
	:param negatives: Negative training sequences.
	:param windowSize: Window size.
	:param windowStep: Window step size.
	:param nTrees: Number of trees.
	:param maxDepth: Maximum tree depth.
	
	:type name: str
	:type features: features
	:type positives: sequences
	:type negatives: sequences
	:type windowSize: int
	:type windowStep: int
	:type nTrees: int
	:type maxDepth: int
	"""
	
	def __init__(self, name, features, trainingSet, windowSize, windowStep, nTrees = 100, maxDepth = None, scale = True, labelPositive = positive, labelNegative = negative):
		super().__init__(name)
		self.windowSize, self.windowStep = windowSize, windowStep
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.trainingSet = trainingSet
		positives, negatives = trainingSet.withLabel([ labelPositive, labelNegative ])
		wPos = sequences(positives.name, [ w for s in positives for w in s.windows(self.windowSize, self.windowStep) ])
		wNeg = sequences(negatives.name, [ w for s in negatives for w in s.windows(self.windowSize, self.windowStep) ])
		if scale:
			features = featureScaler( features, trainingSet = wPos + wNeg )
		self.scale = scale
		self.features = features
		self.nTrees, self.maxDepth = nTrees, maxDepth
		self.threshold = 0.0
		vP = [ self.getSequenceFeatureVector(w) for w in wPos ]
		vN = [ self.getSequenceFeatureVector(w) for w in wNeg ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		self.cls = RandomForestClassifier(n_estimators = nTrees, max_depth = maxDepth, random_state = 0)
		self.cls.fit( np.array(vP+vN), np.array(cP+cN) )
	
	def getTrainer(self):
		return lambda ts: sequenceModelRF(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, nTrees = self.nTrees, maxDepth = self.maxDepth, scale = self.scale, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
	
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	
	def scoreWindow(self, seq):
		return float(self.cls.predict_proba(np.array([self.getSequenceFeatureVector(seq)]))[0][1])
	
	def __str__(self):
		return 'Random Forest<Features: %s; Training set: %s; Positive label: %s; Negative label: %s; Trees: %d; Max. depth: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegatives), self.nTrees, str(self.maxDepth))
	
	def __repr__(self): return self.__str__()

# Lasso
class sequenceModelLasso(sequenceModel):
	"""
	The `sequenceModelLasso` class trains a Lasso model using scikit-learn.
	
	:param name: Model name.
	:param features: Feature set.
	:param positives: Positive training sequences.
	:param negatives: Negative training sequences.
	:param windowSize: Window size.
	:param windowStep: Window step size.
	:param alpha: Alpha parameter for Lasso.
	
	:type name: str
	:type features: features
	:type positives: sequences
	:type negatives: sequences
	:type windowSize: int
	:type windowStep: int
	:type alpha: float
	"""
	
	def __init__(self, name, features, trainingSet, windowSize, windowStep, alpha = 1., labelPositive = positive, labelNegative = negative):
		super().__init__(name)
		self.windowSize, self.windowStep = windowSize, windowStep
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.trainingSet = trainingSet
		positives, negatives = trainingSet.withLabel([ labelPositive, labelNegative ])
		wPos = sequences(positives.name, [ w for s in positives for w in s.windows(self.windowSize, self.windowStep) ])
		wNeg = sequences(negatives.name, [ w for s in negatives for w in s.windows(self.windowSize, self.windowStep) ])
		self.features = features
		self.threshold = 0.0
		vP = [ self.getSequenceFeatureVector(w) for w in wPos ]
		vN = [ self.getSequenceFeatureVector(w) for w in wNeg ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		self.alpha = alpha
		self.cls = Lasso(alpha = alpha)
		self.cls.fit( np.array(vP+vN), np.array(cP+cN) )
	
	def getTrainer(self):
		return lambda ts: sequenceModelLasso(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, alpha = self.alpha, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
	
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	
	def scoreWindow(self, seq):
		return self.cls.predict(np.array([self.getSequenceFeatureVector(seq)]))[0]
	
	def __str__(self):
		return 'Lasso<Features: %s; Training set: %s; Positive label: %s; Negative label: %s; Alpha: %f>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative), self.alpha)
	
	def __repr__(self): return self.__str__()

# Support Vector Machines - Base model
class SVM(baseModel):
	
	def __init__(self, model = None, labelPositive = positive, labelNegative = negative, kDegree = 1, C = 4):
		super().__init__()
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.labelNegative = labelNegative
		self.C, self.kDegree = C, kDegree
		self.mdl = model
	
	def __str__(self):
		return 'Support Vector Machine<C: %s; Kernel degree: %s; Positive label: %s; Negative label: %s>'%(str(self.C), str(self.kDegree), str(self.labelPositive), str(self.labelNegative))
	
	def score(self, featureVectors):
		return [
			[ self.mdl.decision_function(np.array([fv]))[0] ]
			for fv in featureVectors
		]
	
	def train(self, trainingSet):
		fvPos = trainingSet[self.labelPositive]
		fvNeg = trainingSet[self.labelNegative]
		cP = [ 1.0 for _ in range(len(fvPos)) ]
		cN = [ -1.0 for _ in range(len(fvNeg)) ]
		mdl = svm.SVC(C = self.C, kernel = 'poly', degree = self.kDegree, gamma = 'auto')
		mdl.fit( np.array(fvPos+fvNeg), np.array(cP+cN) )
		return SVM(
			model = mdl,
			labelPositive = self.labelPositive,
			labelNegative = self.labelNegative,
			C = self.C,
			kDegree = self.kDegree)
	
	def weights(self, featureNames):
		return nctable(
			'Weights: ' + str(self),
			{
				**{
					'Feature': self.featureNames
				},
				**{
					'Weight': self._weights
				}
			},
			align = { 'Feature': 'l' }
		)
	
	def __repr__(self): return self.__str__()

# Random Forest - Base model
class RF(baseModel):
	
	def __init__(self, model = None, labelPositive = positive, labelNegative = negative, nTrees = 100, maxDepth = None):
		super().__init__()
		self.labelPositive, self.labelNegative = labelPositive, labelNegative
		self.labelNegative = labelNegative
		self.nTrees, self.maxDepth = nTrees, maxDepth
		self.mdl = model
	
	def __str__(self):
		return 'Random Forest<Trees: %s; Max. depth: %s; Positive label: %s; Negative label: %s>'%(str(self.nTrees), str(self.maxDepth), str(self.labelPositive), str(self.labelNegative))
	
	def score(self, featureVectors):
		return [
			#[ self.mdl.decision_function(np.array([fv]))[0] ]
			[ float(self.mdl.predict_proba(np.array([fv]))[0][1]) ]
			for fv in featureVectors
		]
	
	def train(self, trainingSet):
		fvPos = trainingSet[self.labelPositive]
		fvNeg = trainingSet[self.labelNegative]
		cP = [ 1.0 for _ in range(len(fvPos)) ]
		cN = [ -1.0 for _ in range(len(fvNeg)) ]
		#mdl = svm.SVC(C = self.C, kernel = 'poly', degree = self.kDegree, gamma = 'auto')
		mdl = RandomForestClassifier(n_estimators = self.nTrees, max_depth = self.maxDepth, random_state = 0)
		mdl.fit( np.array(fvPos+fvNeg), np.array(cP+cN) )
		return RF(
			model = mdl,
			labelPositive = self.labelPositive,
			labelNegative = self.labelNegative,
			nTrees = self.nTrees,
			maxDepth = self.maxDepth)
	
	def weights(self, featureNames):
		return None
	
	def __repr__(self): return self.__str__()





#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################
# Interfacing with MOCCA

from .regions import loadGFF
from .sequences import sequences, sequenceStream, positive, negative, MarkovChain, IID
from .models import sequenceModel
from os import system

class SVMMOCCA(sequenceModel):
	def __init__(self, name, MOCCAPath, motifs, windowSize, windowStep, kDegree, labelsPositive = [ positive ], labelsNegative = [ negative ], trainingSet = None, silent = True, TempPath = '/dev/shm/mocca/'):
		super().__init__(name, enableMultiprocessing = False)
		self.MOCCAPath = MOCCAPath
		self.tmpPath = TempPath
		self.motifs = motifs
		self.kDegree = kDegree
		self.threshold = 0.0
		self.labelsPositive, self.labelsNegative = labelsPositive, labelsNegative
		self.windowSize, self.windowStep = windowSize, windowStep
		self.silent = silent
		
		# Create temporary files
		system("if [ -e \"" + self.tmpPath + "\" ]; then rm -r " + self.tmpPath + "; fi")
		system('mkdir ' + self.tmpPath)
		labelFiles = {
			lbl: self.tmpPath + 'train_' + lbl.name.replace(' ', '_') + '.fa'
			for lbl in labelsPositive + labelsNegative
		}
		if trainingSet is not None:
			for lbl in labelFiles:
				trainingSet.withLabel(lbl).saveFASTA(labelFiles[lbl])
		
		# Build call path
		self.cpath = '' + MOCCAPath
		self.cpath += ' -C:SVM-MOCCA'
		self.cpath += ' -f:MOCCA:nOcc'
		self.cpath += ' -f:MOCCA:DNT'
		self.cpath += ' -validate:no'
		self.cpath += ' -wSize %d'%self.windowSize
		self.cpath += ' -wStep %d'%self.windowStep
		self.cpath += ' -k:%s'%[ 'linear', 'quadratic', 'cubic' ][ int(kDegree)-1 ]
		self.cpath += ' '
		self.cpath += ' '.join( '-motif:IUPAC "%s" %s %d'%(m.name, m.motif, m.nmismatches) for m in motifs )
		self.cpath += ' '
		self.cpath += ' '.join( '-class "%s" %d +'%(lbl.name, i+1) for i, lbl in enumerate(labelsPositive) )
		self.cpath += ' '
		self.cpath += ' '.join( '-class "%s" -%d -'%(lbl.name, i+1) for i, lbl in enumerate(labelsNegative) )
		self.cpath += ' '
		self.cpath += ' '.join( '-train:FASTA %s %d full'%(labelFiles[lbl], i+1) for i, lbl in enumerate(labelsPositive) )
		self.cpath += ' '
		self.cpath += ' '.join( '-train:FASTA %s -%d full'%(labelFiles[lbl], i+1) for i, lbl in enumerate(labelsNegative) )
	
	def getSequenceScores(self, seqs, cache = True):
		if isinstance(seqs, sequenceStream):
			return self.scoreSequences(seqs.streamFullSequences())
		elif isinstance(seqs, sequences) or isinstance(seqs, list):
			return self.scoreSequences(seqs)
	
	def scoreWindow(self, seq):
		tPath = self.tmpPath + 'vSeq.fasta'
		sequences('', [seq]).saveFASTA(tPath)
		oPath = tPath + '.scores'
		cpath = self.cpath[:] + ' -validate:FASTA ' + tPath + ' + -validate:outSCTable ' + oPath
		if not self.silent:
			print('Debug: SVM-MOCCA - Classifier call: ' + self.cpath)
		if self.silent:
			cpath += ' > ' + self.tmpPath + 'log.txt'
		system(cpath)
		with open(oPath, 'r') as f:
			_ = next(f)
			scores = [ float(p[0]) for p in ( l.strip().split('\t') for l in f ) if len(p) == 2 ]
		return max(scores) if len(scores) > 0 else 0.0
	
	def scoreSequence(self, seq):
		tPath = self.tmpPath + 'vSeq.fasta'
		sequences('', [seq]).saveFASTA(tPath)
		oPath = tPath + '.scores'
		cpath = self.cpath[:] + ' -validate:FASTA ' + tPath + ' + -validate:outSCTable ' + oPath
		if not self.silent:
			print('Debug: SVM-MOCCA - Classifier call: ' + self.cpath)
		if self.silent:
			cpath += ' > ' + self.tmpPath + 'log.txt'
		system(cpath)
		with open(oPath, 'r') as f:
			_ = next(f)
			scores = [ float(p[0]) for p in ( l.strip().split('\t') for l in f ) if len(p) == 2 ]
		score = max(scores) if len(scores) > 0 else 0.0
		return score
	
	def scoreSequences(self, seq):
		tPath = self.tmpPath + 'vSeq.fasta'
		sseq = sequences('', [ s for s in seq ])
		sseq.saveFASTA(tPath)
		oPath = tPath + '.scores'
		cpath = self.cpath[:] + ' -validate:FASTA ' + tPath + ' + -validate:outSCTable ' + oPath
		if not self.silent:
			print('Debug: SVM-MOCCA - Classifier call: ' + self.cpath)
		if self.silent:
			cpath += ' > ' + self.tmpPath + 'log.txt'
		system(cpath)
		with open(oPath, 'r') as f:
			_ = next(f)
			scores = [ float(p[0]) for p in ( l.strip().split('\t') for l in f ) if len(p) == 2 ]
		return scores
		
	def predictSequenceStreamRegions(self, stream):
		pred = regions('Predictions', [])
		windowSP = list(streamSequenceWindows(stream, self.windowSize, self.windowStep))
		windowSequences = sequences('', [ winSeq for winSeq in windowSP ])
		windowPositions = [ winSeq.sourceRegion for winSeq in windowSP ]
		wScores = self.scoreSequences(windowSequences)
		for wScore, winPos in zip(wScores, windowPositions):
			if wScore <= self.threshold: continue
			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:
				cScores = [ ]
				pred.regions.append(winPos)
			else:
				pred[-1].end = winPos.end
			cScores.append(wScore)
			pred[-1].score = sum(cScores) / len(cScores)
		pred.sort()
		return pred
	
	def predictCore(self, genome):
		cmd = self.cpath + ''
		
		cmd += ' -threshold ' + str(self.threshold)
		
		# Run
		cmd += ' -predict:core'
		outGFFPath = self.tmpPath + 'pred.gff'
		cmd += ' -predict:GFF ' + outGFFPath
		genomePath = self.tmpPath + 'genome.fa'
		sequences('', [ s for s in genome.sequences.streamFullSequences() ]).saveFASTA(genomePath)
		cmd += ' -genome:FASTA ' + genomePath
		if self.silent:
			cmd += ' > ' + self.tmpPath + 'log.txt'
		system(cmd)
		return loadGFF(outGFFPath)
	
	def getTrainer(self):
		return lambda ts: SVMMOCCA(name = self.name, MOCCAPath = self.MOCCAPath, motifs = self.motifs, windowSize = self.windowSize, windowStep = self.windowStep, kDegree = self.kDegree, labelsPositive = self.labelsPositive, labelsNegative = self.labelsNegative, trainingSet = ts, silent = self.silent, TempPath = self.tmpPath)
	
	def __str__(self):
		return 'SVM-MOCCA<Motifs: %s; kDegree: %d>'%(str(self.motifs), self.kDegree)
	
	def __repr__(self): return self.__str__()




# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################

progressTasks = []

class progressTask:
	
	def __init__(self, desc, steps):
		self.desc = desc
		self.step = 0
		self.steps = steps
		self.lastDisplay = ''
		progressTasks.append(self)
	
	def update(self, step):
		self.step = step
		self.display()
	
	def display(self, pLen = 20):
		xA = 0.
		xB = 1.
		desc = ''
		#
		for t in progressTasks:
			stepSize = (xB - xA) / t.steps
			cxA = xA + stepSize * t.step
			xA, xB = cxA, cxA + stepSize
			desc += ' - %s (%d/%d)'%(t.desc, t.step+1, t.steps)
		#
		pI = int(xB * pLen)
		pL = pLen - pI
		perc = 100. * xB
		bar = '[' + ('|'*pI) + (' '*pL) + '] %5.2f %%'%perc + desc
		self.lastDisplay = bar
		print(bar, end = '\r')
	
	def done(self):
		print(' ' * len(self.lastDisplay), end = '\r')
		global progressTasks
		progressTasks = [ t for t in progressTasks if t != self ]

def _tblFieldConv(v, fpdec = 5, strcroplen = 16):
	if isinstance(v, int):
		return str(v)
	if isinstance(v, float):
		return ('%%.%df'%fpdec)%v
	v = str(v)
	if strcroplen > 0 and len(v) > strcroplen:
		return v[:int(strcroplen/2)-2] + '...' + v[-int(strcroplen/2)+1:]
	return v

class nctable:
	"""
	The `nctable` class represents a data table. This data table implements a selection of table manipulation methods, and supports formatted, cropped output in text or to HTML for IPython. In addition, the table supports conversion to Numpy arrays and Pandas DataFrames, in order to facilitate wider use cases.
	
	:param title: Title of the table.
	:param _dict: Data in a dictionary. Each key corresponds to a table column.
	:param align: Dictionary of alignment options, from the same keys as in _dict to an alignment option string per column. Use 'l' for left-alignment, or 'r' for right-alignment. If not set, all columns default to right alignment.
	:param ntop: Number of rows to show on the top of the table.
	:param nbottom: Number of rows to show on the bottom of the table.
	:param maxhorz: Maximum number of columns to show.
	:param spaces: Number of spaces to add between columns.
	:param fpdec: Number of decimals to include for floating point numbers.
	:param strcroplen: Maximum string length.
	:param indexName: Name of index column.
	
	:type title: str
	:type _dict: dict
	:type align: dict, optional
	:type ntop: int, optional
	:type nbottom: int, optional
	:type maxhorz: int, optional
	:type spaces: int, optional
	:type fpdec: int, optional
	:type strcroplen: int, optional
	:type indexName: str, optional
	"""
	
	def __init__(self, title, _dict, align = None, ntop = 5, nbottom = 5, maxhorz = 9, spaces = 3, fpdec = 4, strcroplen = 24, indexName = '__index'):
		self.title = title
		if isinstance(_dict, list):
			# Enables table construction by supplying a list of dicts instead of a base dict
			# The nesting is to ensure that order is preserved
			table = _dict
			_dict = {  }
			for e in table:
				for k in e:
					if k not in _dict:
						_dict[k] = [ x[k] if k in x else '' for x in table ]
		self.ntop, self.nbottom = ntop, nbottom
		self.maxhorz = maxhorz
		self.spaces = spaces
		self.align = align
		self.indexName = indexName
		self.nrows = max([len(_dict[k]) for k in _dict if k != self.indexName])
		self.fpdec = fpdec
		self.strcroplen = strcroplen
		if indexName not in _dict.keys():
			self._dict = {
				**{
					self.indexName: list(range(self.nrows))
				},
				**_dict
			}
		else:
			self._dict = _dict
	
	def full(self):
		"""
		Returns a clone of the table, with cropping disabled.
		
		:return: Table without cropping
		:rtype: nctable
		"""
		return nctable(self.title, self._dict, align = self.align, ntop = 0, nbottom = 0, maxhorz = 0, spaces = self.spaces, strcroplen = 0, indexName = self.indexName)
	
	def noCropNames(self):
		"""
		Returns a clone of the table, with cropping of text disabled.
		
		:return: Table without cropping of text
		:rtype: nctable
		"""
		return nctable(self.title, self._dict, align = self.align, ntop = self.ntop, nbottom = self.nbottom, maxhorz = self.maxhorz, spaces = self.spaces, strcroplen = 0, indexName = self.indexName)
	
	def sort(self, key, ascending = True):
		"""
		Returns a clone sorded by a column.
		
		:param key: Column name to sort by.
		:param ascending: If True, then the table will be sorted in ascending order. Otherwise, descending.
		
		:type key: str
		:type ascending: bool, optional
		
		:return: Sorted table
		:rtype: nctable
		"""
		order = sorted(
				[
					i for i, v in enumerate(self._dict[key])
					if isinstance(v, float) or isinstance(v, int)
				],
				key = lambda x: self._dict[key][x]
			) + sorted(
				[
					i for i, v in enumerate(self._dict[key])
					if not (isinstance(v, float) or isinstance(v, int))
				],
				key = lambda x: self._dict[key][x]
			)
		if not ascending: order = list(reversed(order))
		_dict = {
			k: [
				self._dict[k][i]
				for i in order
			]
			for k in self._dict
		}
		return nctable(self.title, _dict, align = self.align, ntop = self.ntop, nbottom = self.nbottom, spaces = self.spaces, strcroplen = self.strcroplen, indexName = self.indexName)
	
	def drop(self, fields):
		"""
		Returns a clone with columns dropped.
		
		:param fields: List of columns to drop. Alternatively, a single string with a column name can be given.
		
		:type fields: str, list
		
		:return: Table with columns dropped
		:rtype: nctable
		"""
		if isinstance(fields, str):
			fields = [ fields ]
		_dict = {
			k: self._dict[k]
			for k in self._dict
			if k not in fields
		}
		return nctable(self.title, _dict, align = self.align, ntop = self.ntop, nbottom = self.nbottom, spaces = self.spaces, strcroplen = self.strcroplen, indexName = self.indexName)
	
	def summary(self):
		"""
		Returns a summary table, with means, variances and standard deviations of each field.
		
		:return: Summary table
		:rtype: nctable
		"""
		src = self._dict
		keys = [ k for k in src if k != self.indexName ]
		isnum = {
			k: all(isinstance(x, float) or isinstance(x, int) for x in src[k])
			for k in keys
		}
		mean = {
			k: sum(src[k]) / len(src[k]) if isnum[k] else 'N/A'
			for k in keys
		}
		var = {
			k: sum((v - mean[k])**2. for v in src[k]) if isnum[k] else 'N/A'
			for k in keys
		}
		std = {
			k: ((var[k]/(len(src[k])-1))**0.5 if len(src[k]) > 1 else 0) if isnum[k] else 'N/A'
			for k in keys
		}
		_dict = {
			'Field': [ k for k in keys ],
			'Type': [ 'Number' if isnum[k] else 'Text' for k in keys ],
			'Mean': [ mean[k] for k in keys ],
			'Var.': [ var[k] for k in keys ],
			'St.d.': [ std[k] for k in keys ],
		}
		return nctable('Summary table: ' + self.title, _dict, align = { 'Field': 'l' })
	
	def __iter__(self):
		return self._dict.__iter__()
	
	def __getitem__(self, i):
		return self._dict[i]
	
	def __array__(self):
		try:
			import numpy as np
			values = {}
			types = {}
			for k in self._dict:
				if k == self.indexName: continue
				if all(isinstance(v, float) for v in self._dict[k]):
					values[k] = self._dict[k] + [
						np.na for _ in range(len(self._dict[k]), self.nrows) ]
					types[k] = 'f8'
				elif all(isinstance(v, int) for v in self._dict[k]):
					values[k] = self._dict[k] + [
						np.na for _ in range(len(self._dict[k]), self.nrows) ]
					types[k] = 'i4'
				else:
					values[k] = [ str(v) for v in self._dict[k] ] + [
						'' for _ in range(len(self._dict[k]), self.nrows) ]
					types[k] = 'a%d'%( max( len(v) for v in values[k] ) + 1 )
			return np.array([
					tuple([
						values[k][i]
						for k in self._dict if k != self.indexName
					])
					for i in range(self.nrows)
				],
				dtype=[
					(k, types[k])
					for k in self._dict if k != self.indexName
				])
		except ImportError as err:
			raise err
	
	def to_numpy(self):
		"""
		Returns a Numpy array generated from the table.
		
		:return: Numpy array
		:rtype: numpy.array
		"""
		return self.__array__()
	
	def to_pandas(self):
		"""
		Returns a Pandas DataFrame generated from the table.
		
		:return: Pandas DataFrame
		:rtype: pandas.DataFrame
		"""
		try:
			import pandas as pd
			return pd.DataFrame(self.__array__())
		except ImportError as err:
			raise err
	
	def _cropv(self):
		_dict, ntop, nbottom = self._dict, self.ntop, self.nbottom
		# Horizontal cropping
		keys = [ k for k in _dict ]
		if len(keys) > self.maxhorz and self.maxhorz != 0:
			_dict = {
				**{
					k: _dict[k]
					for k in keys[:int(self.maxhorz/2)]
				},
				**{
					'...': [ '...' for _ in range(max(len(_dict[k]) for k in _dict)) ]
				},
				**{
					k: _dict[k]
					for k in keys[-int(self.maxhorz/2):]
				},
			}
		# Vertical cropping
		if ntop == 0 and nbottom == 0:
			return _dict
		if self.nrows <= ntop + nbottom:
			return _dict
		rdict = { k: [] for k in _dict }
		if ntop > 0:
			for i in range(ntop):
				for k in _dict:
					rdict[k].append(_dict[k][i] if i < len(_dict[k]) else '')
			if nbottom > 0:
				for k in _dict:
					rdict[k].append('...')
		if nbottom > 0:
			for i in range(self.nrows-nbottom, self.nrows):
				for k in _dict:
					rdict[k].append(_dict[k][i] if i < len(_dict[k]) else '')
		return rdict
	
	def _prepv(self):
		_dict = self._cropv()
		return {
			_tblFieldConv(k, strcroplen = self.strcroplen): [
				_tblFieldConv(v, fpdec = self.fpdec, strcroplen = self.strcroplen)
				for v in _dict[k]
			]
			for k in _dict
		}
	
	def __repr__(self):
		sp = ''.join(' ' for _ in range(self.spaces))
		_dict = self._prepv()
		align = {
			k: self.align[k]
			if ((self.align is not None) and k in self.align.keys())
			else 'r'
			for k in _dict
		}
		fmt = {
			k: ('%%-%ds' if align[k] == 'l' else '%%%ds')%max([
				0 if k == self.indexName else len(k)
			] + [
				len(e) for e in _dict[k]
			])
			for k in _dict
		}
		hdr = sp.join(
			fmt[k]%('' if k == self.indexName else k)
			for k in _dict
		)
		nrows = max([len(_dict[k]) for k in _dict if k != self.indexName])
		body = '\n'.join(
			sp.join(
				fmt[k]%(_dict[k][i] if i < len(_dict[k]) else '')
				for k in _dict
			)
			for i in range(nrows)
		)
		return self.title + '\n' +\
			'Rows: %d\n'%self.nrows +\
			'Columns: %d\n'%len([ 1 for k in self._dict if k != self.indexName ]) +\
			hdr + '\n' + body
	
	def _repr_html_(self):
		_dict = self._prepv()
		sec = lambda x: x.replace('<', '&lt;').replace('>', '&gt;')
		hdr = '<thead><th></th><th>' + '</th><th>'.join(
			'<b>%s</b>'%str(sec(k))
			for k in _dict if k != self.indexName
		) + '</th></thead>'
		nrows = max([len(_dict[k]) for k in _dict if k != self.indexName])
		body = '<tbody><tr>' + '</tr><tr>'.join(
			'<td><b>%s</b></td>'%_dict[self.indexName][i] + '<td>' + '</td><td>'.join(
				sec(_dict[k][i]) if i < len(_dict[k]) else ''
				for k in _dict if k != self.indexName
			) + '</td>'
			for i in range(nrows)
		) + '</tr></tbody>'
		return '<div><div>' + sec(self.title) + '</div>' +\
			'<div>Rows: %d</div>'%self.nrows +\
			'<div>Columns: %d</div>'%len([ 1 for k in self._dict if k != self.indexName ]) +\
			'<table>' +\
			hdr + body + '</table></div>'




#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################

from setuptools import setup
from setuptools.extension import Extension

setup(
	ext_modules = [ Extension(name = 'gnocis.' + x, sources = ['gnocis/'+ x + '.cpp']) for x in ['biomarkers', 'common', 'features', 'featurenetwork', 'models', 'motifs', 'regions', 'curves', 'sequences', 'validation'] ],
	name = "gnocis",
	packages = [ "gnocis" ],
	author = "Bjørn André Bredesen-Aa",
	author_email = "bjorn@bjornbredesen.no",
	version = "0.9.12",
	url = "https://github.com/bjornbredesen/gnocis",
	license = "MIT",
	description = "Gnocis is a system for the analysis and the modelling of cis-regulatory DNA sequences.",
	long_description = open("README.md", encoding="utf-8").read(),
	long_description_content_type = "text/markdown",
	classifiers=[
		"Development Status :: 4 - Beta",
		"Environment :: Console",
		"Intended Audience :: Developers",
		"Intended Audience :: Information Technology",
		"License :: OSI Approved :: MIT License",
		"Operating System :: POSIX :: Linux",
		"Operating System :: Microsoft :: Windows",
		"Operating System :: MacOS",
		"Programming Language :: Cython",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Topic :: Scientific/Engineering",
		"Topic :: Scientific/Engineering :: Bio-Informatics",
	],
)

